// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	DOOM graphics stuff for X11, UNIX.
//
//-----------------------------------------------------------------------------

static const char
rcsid[] = "$Id: i_x.c,v 1.6 1997/02/03 22:45:10 b1 Exp $";

#pragma hdrstop

#include <stdlib.h>
#include <assert.h>

#ifdef USE_X11
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <X11/extensions/XShm.h>
// Had to dig up XShm.c for this one.
// It is in the libXext, but not in the XFree86 headers.
#ifdef LINUX
int XShmGetEventBase( Display* dpy ); // problems with g++?

//bk20050214 - #include <errnos.h>
#include <errno.h>
#else
#include <errnos.h>
#endif

#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>

#include <signal.h>
#endif

#include "info.h"
#include "doomstat.h"
#include "i_system.h"
#include "v_video.h"
#include "m_argv.h"
#include "d_main.h"

#include "doomdef.h"

#ifndef USE_X11
#include "../TerminalDoom.h"
#endif

#define POINTER_WARP_COUNTDOWN	1

unsigned char*	debug_screenEnd = NULL; // debug

#ifdef USE_X11
Display*	X_display=0;
Window		X_mainWindow;

Visual*		X_visual;
GC		X_gc;
XEvent		X_event;
int		X_screen;
XVisualInfo	X_visualinfo;
XImage*		image;
int		X_width;
int		X_height;

// MIT SHared Memory extension.
boolean		doShm;

XShmSegmentInfo	X_shminfo;
int		X_shmeventtype;
#else // USE_X11
void*           ptrLFB = NULL;
#endif // USE_X11

// Fake mouse handling.
// This cannot work properly w/o DGA.
// Needs an invisible mouse cursor at least.
boolean		grabMouse;
int		doPointerWarp = POINTER_WARP_COUNTDOWN;

// Blocky mode,
// replace each 320x200 pixel with multiply*multiply pixels.
// According to Dave Taylor, it still is a bonehead thing
// to use ....
// bk - we do not multiply
//static int	multiply=1;

// Palette stuff.
unsigned int colors[256];

//
//  Translates the key currently in X_event
//

int xlatekey(void)
{
    int rc = 0;
#ifdef USE_X11
    switch(rc = XKeycodeToKeysym(X_display, X_event.xkey.keycode, 0))
    {
      case XK_Left:	rc = KEY_LEFTARROW;	break;
      case XK_Right:	rc = KEY_RIGHTARROW;	break;
      case XK_Down:	rc = KEY_DOWNARROW;	break;
      case XK_Up:	rc = KEY_UPARROW;	break;
      case XK_Escape:	rc = KEY_ESCAPE;	break;
      case XK_Return:	rc = KEY_ENTER;		break;
      case XK_Tab:	rc = KEY_TAB;		break;
      case XK_F1:	rc = KEY_F1;		break;
      case XK_F2:	rc = KEY_F2;		break;
      case XK_F3:	rc = KEY_F3;		break;
      case XK_F4:	rc = KEY_F4;		break;
      case XK_F5:	rc = KEY_F5;		break;
      case XK_F6:	rc = KEY_F6;		break;
      case XK_F7:	rc = KEY_F7;		break;
      case XK_F8:	rc = KEY_F8;		break;
      case XK_F9:	rc = KEY_F9;		break;
      case XK_F10:	rc = KEY_F10;		break;
      case XK_F11:	rc = KEY_F11;		break;
      case XK_F12:	rc = KEY_F12;		break;
	
      case XK_BackSpace:
      case XK_Delete:	rc = KEY_BACKSPACE;	break;

      case XK_Pause:	rc = KEY_PAUSE;		break;

      case XK_KP_Equal:
      case XK_equal:	rc = KEY_EQUALS;	break;

      case XK_KP_Subtract:
      case XK_minus:	rc = KEY_MINUS;		break;

      case XK_Shift_L:
      case XK_Shift_R:
	rc = KEY_RSHIFT;
	break;
	
      case XK_Control_L:
      case XK_Control_R:
	rc = KEY_RCTRL;
	break;
	
      case XK_Alt_L:
      case XK_Meta_L:
      case XK_Alt_R:
      case XK_Meta_R:
	rc = KEY_RALT;
	break;
	
      default:
	if (rc >= XK_space && rc <= XK_asciitilde)
	    rc = rc - XK_space + ' ';
	if (rc >= 'A' && rc <= 'Z')
	    rc = rc - 'A' + 'a';
	break;
    }
#endif
    return rc;

}


void I_ShutdownGraphics(void)
{
#ifdef USE_X11
  // Detach from X server
  if (!XShmDetach(X_display, &X_shminfo))
	    I_Error("XShmDetach() failed in I_ShutdownGraphics()");

  // Release shared memory.
  shmdt(X_shminfo.shmaddr);
  shmctl(X_shminfo.shmid, IPC_RMID, 0);

  // Paranoia.
  image->data = NULL;
#endif
}



//
// I_StartFrame
//
void I_StartFrame (void)
{
    // er?

}

#ifdef USE_X11
static int	lastmousex = 0;
static int	lastmousey = 0;
#endif

boolean		mousemoved = false;
boolean		shmFinished;

void I_GetEvent(void)
{
    event_t event;

#ifdef USE_X11
    // put event-grabbing stuff in here
    XNextEvent(X_display, &X_event);
    switch (X_event.type)
    {
      case KeyPress:
		event.type = ev_keydown;
		event.data1 = xlatekey();
		D_PostEvent(&event);
		// fprintf(stderr, "k");
		break;
      case KeyRelease:
		event.type = ev_keyup;
		event.data1 = xlatekey();
		D_PostEvent(&event);
		// fprintf(stderr, "ku");
	break;
      case ButtonPress:
		event.type = ev_mouse;
		event.data1 =
			(X_event.xbutton.state & Button1Mask)
			| (X_event.xbutton.state & Button2Mask ? 2 : 0)
			| (X_event.xbutton.state & Button3Mask ? 4 : 0)
			| (X_event.xbutton.button == Button1)
			| (X_event.xbutton.button == Button2 ? 2 : 0)
			| (X_event.xbutton.button == Button3 ? 4 : 0);
		event.data2 = event.data3 = 0;
		D_PostEvent(&event);
		// fprintf(stderr, "b");
	break;
      case ButtonRelease:
		event.type = ev_mouse;
		event.data1 =
			(X_event.xbutton.state & Button1Mask)
			| (X_event.xbutton.state & Button2Mask ? 2 : 0)
			| (X_event.xbutton.state & Button3Mask ? 4 : 0);
		// suggest parentheses around arithmetic in operand of |
		event.data1 =
			event.data1
			 ^ (X_event.xbutton.button == Button1 ? 1 : 0)
			 ^ (X_event.xbutton.button == Button2 ? 2 : 0)
			^ (X_event.xbutton.button == Button3 ? 4 : 0);
		event.data2 = event.data3 = 0;
		D_PostEvent(&event);
		// fprintf(stderr, "bu");
		break;
      case MotionNotify:
		event.type = ev_mouse;
		event.data1 =
			(X_event.xmotion.state & Button1Mask)
			| (X_event.xmotion.state & Button2Mask ? 2 : 0)
			| (X_event.xmotion.state & Button3Mask ? 4 : 0);
		event.data2 = (X_event.xmotion.x - lastmousex) << 2;
		event.data3 = (lastmousey - X_event.xmotion.y) << 2;

		if (event.data2 || event.data3)
		{
		 lastmousex = X_event.xmotion.x;
		 lastmousey = X_event.xmotion.y;
		 if (X_event.xmotion.x != X_width/2 &&
			X_event.xmotion.y != X_height/2)
			{
				D_PostEvent(&event);
				// fprintf(stderr, "m");
				mousemoved = false;
			} else
			{
				mousemoved = false;
		 }
		}
		break;
	
      case Expose:
      case ConfigureNotify:
		break;
	
      default:
		if (doShm && X_event.type == X_shmeventtype) shmFinished = true;
		break;
    }
#else
    int key = 0;
    bool up = false;

    TheTerminal::GenerateEvents();
    while ( TheTerminal::NextEvent(&key,&up) )
    {
		event.type = (up? ev_keyup:ev_keydown);
		event.data1 = key;
		event.data2 = event.data3 = 0;
		D_PostEvent(&event);
    }
    int deltaX = 0;
    int deltaY = 0;    
	int buttons = 0;
    TheTerminal::GetMouseEvent( &deltaX, &deltaY, &buttons );
	//if ( deltaX!=0 || deltaY!=0 || buttons !=0 )
	{
		event.type = ev_mouse;
		event.data1 = buttons;
		event.data2 = deltaX;
		event.data3 = (-deltaY); // flip sign
		D_PostEvent(&event);
		//if ( buttons ) I_Printf("buttons: 0x%x", buttons );
	}
#endif
}

#ifdef USE_X11
Cursor
createnullcursor
( Display*	display,
  Window	root )
{
    Cursor cursor;


    Pixmap cursormask;
    XGCValues xgc;
    GC gc;
    XColor dummycolour;

    cursormask = XCreatePixmap(display, root, 1, 1, 1/*depth*/);
    xgc.function = GXclear;
    gc =  XCreateGC(display, cursormask, GCFunction, &xgc);
    XFillRectangle(display, cursormask, gc, 0, 0, 1, 1);
    dummycolour.pixel = 0;
    dummycolour.red = 0;
    dummycolour.flags = 04;
    cursor = XCreatePixmapCursor(display, cursormask, cursormask,
				 &dummycolour,&dummycolour, 0,0);
    XFreePixmap(display,cursormask);
    XFreeGC(display,gc);

    return cursor;
}
#endif

//
// I_StartTic
//
void I_StartTic (void)
{
#ifdef USE_X11
    if (!X_display)
	return;

    while (XPending(X_display))
	I_GetEvent();

    // Warp the pointer back to the middle of the window
    //  or it will wander off - that is, the game will
    //  loose input focus within X11.
    if (grabMouse)
    {
	if (!--doPointerWarp)
	{
	    XWarpPointer( X_display,
			  None,
			  X_mainWindow,
			  0, 0,
			  0, 0,
			  X_width/2, X_height/2);

	    doPointerWarp = POINTER_WARP_COUNTDOWN;
	}
    }

    mousemoved = false;
#else
    I_GetEvent();
#endif
}


//
// I_UpdateNoBlit
//
void I_UpdateNoBlit (void)
{
    // what is this?
}


void I_UpdateBlank  ( int scrn )
{
  if ( scrn==-1 )
  {
	I_UpdateBlank  ( 0 );
	I_UpdateBlank  ( 1 );
	I_UpdateBlank  ( 2 );
	I_UpdateBlank  ( 3 );
	I_UpdateBlank  ( 4 );
	return;
  }
  assert(scrn>=0);
  assert(scrn<5);
  memset( screens[scrn], 0x0, SCREENHEIGHT*SCREENWIDTH );
}

#define BUFFERWIDTH  512

//
// I_FinishUpdate
//
void I_FinishUpdate (void)
{

    static int	lasttic;
    int		tics;
    int		i;
    // UNUSED static unsigned char *bigscreen=0;

    // draws little dots on the bottom of the screen
    if (devparm)
    {

	i = I_GetTime();
	tics = i - lasttic;
	lasttic = i;
	if (tics > 20) tics = 20;

	for (i=0 ; i<tics*2 ; i+=2)
	    screens[0][ (SCREENHEIGHT-1)*SCREENWIDTH + i] = 0xff;
	for ( ; i<20*2 ; i+=2)
	    screens[0][ (SCREENHEIGHT-1)*SCREENWIDTH + i] = 0x0;
    
    }

    // scales the screen size before blitting it
    // bk: we blit multiply from 320x200 8bit to RGB(A)

    int y;
    unsigned char* ilineptr = (unsigned char *)(screens[0]);

#ifdef USE_X11 
    unsigned int* olineptr = (unsigned int *) &(image->data[0]);

    for ( y=0; y<SCREENHEIGHT; y++ )
    {
   
      int x;
      for ( x=0; x<SCREENWIDTH; x++ )
      {
	*olineptr = colors[ *ilineptr ];
	ilineptr++;
	olineptr++;
      }
    } 

#else//USE_X11

    if ( ptrLFB==NULL )
    {
      fprintf( stderr, "No LFB!" );
      return;
    }
    

#if 0 // bk: checkerboard test pattern - works!
    int h,w;
    for ( h=0; h<SCREENHEIGHT; h++ )
    {
      for ( w=0; w<SCREENWIDTH; w++ )
      {
	unsigned char* pIndex = (unsigned char *)(screens[0])+(w+(h*SCREENWIDTH));
	if ( ((w>>5)&0x1)!=((h>>5)&0x1))
	{
	  *pIndex = 0;
	}
	else
	{
	  *pIndex = 16;
	}
      }
    } 
#endif

    //fprintf( stderr, "DOOM blit! \n");

    for ( y=0; y<SCREENHEIGHT; y++ )
    {
      unsigned int* olineptr = ((unsigned int *)ptrLFB)+(y*BUFFERWIDTH);
      
      int x;
      for ( x=0; x<SCREENWIDTH; x++ )
      {
	if ( x<BUFFERWIDTH)
	{
	  *olineptr = colors[ *ilineptr ];   // loop works: 0xFF00FF00 gives green 
	}
	ilineptr++;
	olineptr++;
      }
    } 

#endif//USE_X11

    //memset( screens[0], 0x0, SCREENHEIGHT*SCREENWIDTH );

#ifdef USE_X11
    //if (doShm)
    {
	if (!XShmPutImage(	X_display,
				X_mainWindow,
				X_gc,
				image,
				0, 0,
				0, 0,
				X_width, X_height,
				True ))
	    I_Error("XShmPutImage() failed\n");

	// wait for it to finish and processes all input events
	shmFinished = false;
	do
	{
	    I_GetEvent();
	} while (!shmFinished);

    }
#else
    //I_GetEvent();
#endif
}


//
// I_ReadScreen
//
void I_ReadScreen (byte* scr)
{
    memcpy (scr, screens[0], SCREENWIDTH*SCREENHEIGHT);
}


//
// I_SetPalette
//
void I_SetPalette (byte* palette)
{
  // bk : set the colormap entries
  int i;
  for (i=0 ; i<256 ; i++)
  {
    unsigned int red = gammatable[usegamma][*palette++];
    unsigned int green = gammatable[usegamma][*palette++];
    unsigned int blue = gammatable[usegamma][*palette++];

#ifdef USE_X11	

    colors[i] = (red<<16) + (green<<8) + blue;

#else//USE_X11

    const unsigned int alpha = 0xFF;
    colors[i] = (red)+(green<<8)+(blue<<16)+(alpha<<24);

#endif//USE_X11
  }

  //fprintf( stderr, "set palette (0x%x) ! \n", (int)palette );
}


//
// This function is probably redundant,
//  if XShmDetach works properly.
// ddt never detached the XShm memory,
//  thus there might have been stale
//  handles accumulating.
//
void grabsharedmemory( unsigned int size)
{
#ifdef USE_X11
  int			key = ('d'<<24) | ('o'<<16) | ('o'<<8) | 'm';
  struct shmid_ds	shminfo;
  int			minsize = 320*200;
  int			id;
  int			rc;
  // UNUSED int done=0;
  int			pollution=5;
  
  // try to use what was here before
  do
  {
    id = shmget((key_t) key, minsize, 0777); // just get the id
    if (id != -1)
    {
      rc=shmctl(id, IPC_STAT, &shminfo); // get stats on it
      if (!rc) 
      {
	if (shminfo.shm_nattch)
	{
	  fprintf(stderr, "User %d appears to be running "
		  "DOOM.  Is that wise?\n", shminfo.shm_cpid);
	  key++;
	}
	else
	{
	  if (getuid() == shminfo.shm_perm.cuid)
	  {
	    rc = shmctl(id, IPC_RMID, 0);
	    if (!rc)
	      fprintf(stderr,
		      "Was able to kill my old shared memory\n");
	    else
	      I_Error("Was NOT able to kill my old shared memory");
	    
	    id = shmget((key_t)key, size, IPC_CREAT|0777);
	    if (id==-1)
	      I_Error("Could not get shared memory");
	    
	    rc=shmctl(id, IPC_STAT, &shminfo);
	    
	    break;
	    
	  }
	  if ( size >= shminfo.shm_segsz)
	  {
	    fprintf(stderr,
		    "will use %d's stale shared memory\n",
		    shminfo.shm_cpid);
	    break;
	  }
	  else
	  {
	    fprintf(stderr,
		    "warning: can't use stale "
		    "shared memory belonging to id %d, "
		    "key=0x%x\n",
		    shminfo.shm_cpid, key);
	    key++;
	  }
	}
      }
      else
      {
	I_Error("could not get stats on key=%d", key);
      }
    }
    else
    {
      id = shmget((key_t)key, size, IPC_CREAT|0777);
      if (id==-1)
      {
#ifndef __cplusplus
	extern int errno;
	fprintf(stderr, "errno=%d\n", errno);
#endif
	I_Error("Could not get any shared memory");
      }
      break;
    }
  } while (--pollution);
  
  if (!pollution)
  {
    I_Error("Sorry, system too polluted with stale "
	    "shared memory segments.\n");
    }	
  
  X_shminfo.shmid = id;
  
  // attach to the shared memory segment
  X_shminfo.shmaddr = (char*) shmat(id, 0, 0);
  image->data = X_shminfo.shmaddr;

  fprintf(stderr, "shared memory id=%d, addr=0x%x\n", id,
	  (int) (image->data));
#endif
}


void I_InitGraphics( void* pLFB )
{
#ifdef USE_X11
    char*		displayname;
    char*		d;
    int			n;
    int			pnum;
    int			x=0;
    int			y=0;
    
    // warning: char format, different type arg
    char		xsign=' ';
    char		ysign=' ';
    
    int			oktodraw;
    unsigned long	attribmask;
    XSetWindowAttributes attribs;
    XGCValues		xgcvalues;
    int			valuemask;
    static int		firsttime=1;

    if (!firsttime)
	return;
    firsttime = 0;

    signal(SIGINT, (void (*)(int)) I_Quit);
   
    // bk - we render at 320x200
    X_width = SCREENWIDTH;
    X_height = SCREENHEIGHT;

    // check for command-line display name
    if ( (pnum=M_CheckParm("-disp")) ) // suggest parentheses around assignment
	displayname = myargv[pnum+1];
    else
	displayname = 0;

    // check if the user wants to grab the mouse (quite unnice)
    grabMouse = !!M_CheckParm("-grabmouse");

    // check for command-line geometry
    if ( (pnum=M_CheckParm("-geom")) ) // suggest parentheses around assignment
    {
	// warning: char format, different type arg 3,5
	n = sscanf(myargv[pnum+1], "%c%d%c%d", &xsign, &x, &ysign, &y);
	
	if (n==2)
	    x = y = 0;
	else if (n==6)
	{
	    if (xsign == '-')
		x = -x;
	    if (ysign == '-')
		y = -y;
	}
	else
	    I_Error("bad -geom parameter");
    }

    // open the display
    X_display = XOpenDisplay(displayname);
    if (!X_display)
    {
	if (displayname)
	    I_Error("Could not open display [%s]", displayname);
	else
	    I_Error("Could not open display (DISPLAY=[%s])", getenv("DISPLAY"));
    }

    // bk - use the default visual if true color 
    X_screen = DefaultScreen(X_display);
    if (!XMatchVisualInfo(X_display, X_screen, 24, TrueColor, &X_visualinfo))
	I_Error("truedoom currently only supports 24-bit screens");
    X_visual = X_visualinfo.visual;

    // check for the MITSHM extension
    doShm = XShmQueryExtension(X_display);

    if (!doShm)
	I_Error("truedoom currently only supports MIT SHM");    

    // even if it's available, make sure it's a local connection  
    if (doShm)  
    {
	if (!displayname) displayname = (char *) getenv("DISPLAY");
	if (displayname)
	{
	    d = displayname;
	    while (*d && (*d != ':')) d++;
	    if (*d) *d = 0;
	    if (!(!strcasecmp(displayname, "unix") || !*displayname)) doShm = false;
	}
    }


    // bk - SHM only
    if (!doShm)
	I_Error("truedoom currently only supports local MIT SHM");    

    fprintf(stderr, "Using MITSHM extension\n");

    // bk: do not create the colormap
    //X_cmap = XCreateColormap(X_display, RootWindow(X_display,X_screen), X_visual, AllocAll);

    // setup attributes for main window
    attribmask = CWEventMask | /*CWColormap |*/ CWBorderPixel;
    attribs.event_mask =
	KeyPressMask
	| KeyReleaseMask
	// | PointerMotionMask | ButtonPressMask | ButtonReleaseMask
	| ExposureMask;

    //attribs.colormap = X_cmap;
    attribs.border_pixel = 0;

    // create the main window
    X_mainWindow = XCreateWindow(	X_display,
					RootWindow(X_display, X_screen),
					x, y,
					X_width, X_height,
					0, // borderwidth
					24, // bk 8, // depth
					InputOutput,
					X_visual,
					attribmask,
					&attribs );

    XDefineCursor(X_display, X_mainWindow,
		  createnullcursor( X_display, X_mainWindow ) );

    // create the GC
    valuemask = GCGraphicsExposures;
    xgcvalues.graphics_exposures = False;
    X_gc = XCreateGC(	X_display,
  			X_mainWindow,
  			valuemask,
  			&xgcvalues );

    // map the window
    XMapWindow(X_display, X_mainWindow);

    // wait until it is OK to draw
    oktodraw = 0;
    while (!oktodraw)
    {
	XNextEvent(X_display, &X_event);
	if (X_event.type == Expose
	    && !X_event.xexpose.count)
	{
	    oktodraw = 1;
	}
    }

    // grabs the pointer so it is restricted to this window
    if (grabMouse)
	XGrabPointer(X_display, X_mainWindow, True,
		     ButtonPressMask|ButtonReleaseMask|PointerMotionMask,
		     GrabModeAsync, GrabModeAsync,
		     X_mainWindow, None, CurrentTime);

    if (doShm)
    {

	X_shmeventtype = XShmGetEventBase(X_display) + ShmCompletion;

	// create the image
	image = XShmCreateImage(	X_display,
					X_visual,
					24, //8,
					ZPixmap,
					0,
					&X_shminfo,
					X_width,
					X_height );

	grabsharedmemory(image->bytes_per_line * image->height);


	// UNUSED
	// create the shared memory segment
	// X_shminfo.shmid = shmget (IPC_PRIVATE,
	// image->bytes_per_line * image->height, IPC_CREAT | 0777);
	// if (X_shminfo.shmid < 0)
	// {
	// perror("");
	// I_Error("shmget() failed in InitGraphics()");
	// }
	// fprintf(stderr, "shared memory id=%d\n", X_shminfo.shmid);
	// attach to the shared memory segment
	// image->data = X_shminfo.shmaddr = shmat(X_shminfo.shmid, 0, 0);
	

	if (!image->data)
	{
	    perror("");
	    I_Error("shmat() failed in InitGraphics()");
	}

	// get the X server to attach to it
	if (!XShmAttach(X_display, &X_shminfo))
	    I_Error("XShmAttach() failed in InitGraphics()");

    }
    else
    {
            // bk: dead branch
      	    I_Error("XShmAttach() failed in InitGraphics()");
    }

    // bk : NOT  screens[0] = (unsigned char *) (image->data);

#else // USE_X11

    ptrLFB = pLFB; 

#endif // USE_X11

    screens[0] = (unsigned char *) malloc (SCREENWIDTH * SCREENHEIGHT);
	I_UpdateBlank(0);
	debug_screenEnd = screens[0]+((SCREENWIDTH*SCREENHEIGHT)-1); // last pixel
}
